/* 
 *  $Id: Channel.Java,v 1.7 1998/05/29 20:51:46 gback Exp $
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */


package ANTS_PKG;
import NODEOS_PKG.*;

import java.io.*;
import java.net.*;

import UTILS_PKG.*;

/**
 * Channel used by nodes to exchanged capsule with one another.
 * <p>
 * The abstract class <tt>Channel</tt> knows how to encode outgoing capsules.  
 * Subclasses, such as UDPChannel, implement the actual send operation.
 *
 * @see UDPChannel
 *
 * @author David Wetherall
 */

public abstract class Channel
  extends ManagedObject
{
  public ChannelAddress source;
  public Node client;
  public int clientAddress;
  protected Buffer txbuf;

  static byte ANEPVERSION = 1;
  static short ANTSTYPE = 18;
  static short SRCOPTTYPE = 1;
  static short DESTOPTTYPE = 2;
  static int IPV4SCHEME = 1;
  protected static int ANTSVERSION = 0x00010002;

  /**
   * Encode an outgoing capsule.
   * <p>
   * This writes the ANEP and ANTS header for an outgoing capsule in its
   * associated Xdr.
   * 
   * @param cap Capsule to be encoded.
   * @return Xdr returns Xdr in which the rest of the capsule can be encoded.
   */
  public Xdr encode(Capsule cap) {
    short packetLen;

    // source, etc. must be set already
    if (cap.src == 0) {
      Entity.warn("SecurityException: source not set in encode for "+ cap);
      throw new SecurityException();
    }
        
    // previous is updated for use by the code loading
    cap.previous = clientAddress;
 
    // resources decremented here just before we leave
    if (--cap.resources <= 0) {
      throw new ResourceLimitException();
    }

    Xdr xdr;
    
    if (cap.xdr != null) {
      // the capsule is reusing it own buffer - reset it!
      xdr = cap.xdr.reset();
      // SEARCH FOR SOURCE AND DEST OPTIONS AND CHANGE VALUES ACCORDINGLY
      xdr.index += Xdr.INT;
      short headerLen = xdr.SHORT();
      packetLen = (short)(headerLen * Xdr.INT + Xdr.INT + cap.length());
      xdr.len = packetLen;
      xdr.PUT(packetLen);
      while (xdr.index < headerLen * Xdr.INT) {
	short optType = (short)(xdr.SHORT() & 0x3FFF);
	short optLen = xdr.SHORT();
	if (optType == DESTOPTTYPE) {
	  xdr.index += Xdr.INT;   // Skip over scheme identifier
	  xdr.PUT(cap.getDst());
	  break;
	} else {
	  xdr.index += (optLen - 1) * Xdr.INT;
	}
      }
      xdr.index = headerLen * Xdr.INT + Xdr.INT;
    }
    else {
      packetLen = (short)(32 + Xdr.INT + cap.length());
      xdr = new Xdr(txbuf, 0, packetLen);
      // Insert Header
      xdr.PUT(ANEPVERSION);
      xdr.PUT((byte)0);  // Forward packet if TypeID unknown
      xdr.PUT(ANTSTYPE);

      // Header with src and dest options = 8 32-bit words
      xdr.PUT((short)8);
      xdr.PUT(packetLen);

      xdr.PUT(DESTOPTTYPE);
      xdr.PUT((short)3);
      xdr.PUT(IPV4SCHEME);
      xdr.PUT(cap.getDst());

      xdr.PUT(SRCOPTTYPE);
      xdr.PUT((short)3);
      xdr.PUT(IPV4SCHEME);
      xdr.PUT(cap.getSrc());

      xdr.PUT(ANTSVERSION);
    }
    cap.xdr = xdr;
    cap.encode();

    return xdr;
  }

  public void attachNode(Node c) {
    if (client != c) {
      client = c;
      clientAddress = c.getAddress();
      client.attachChannel(this);
    }
  }
  
  /**
   * Send a capsule to a given destination.
   */
  public final boolean send(Capsule cap, ChannelAddress dst) {
    Xdr xdr = encode(cap);
    if (xdr.offset != 0) {
      byte[] bytes = new byte[xdr.len];
      System.arraycopy(xdr.buf, xdr.offset, bytes, 0, xdr.len);
      return send(bytes, xdr.len, dst);
    }
    else
      return send(xdr.buf, xdr.len, dst);
  }

  abstract boolean send(byte[] buf, int length, ChannelAddress dst);

  static public Channel New(Manager b, String n, String src) 
    throws Exception
  {
    ChannelAddress c = ChannelAddress.New(src);
    
    if (c instanceof UDPChannelAddress) {
      return new UDPChannel(b, n, (UDPChannelAddress)c);
    } else {
      throw new Exception("unrecognized type of channel" + src);
    }
  }
  
  public Channel(Manager b, String n, ChannelAddress src) {
    this(b, n, src, 32);
  }

  public Channel(Manager b, String n, ChannelAddress src, int q) {
    super(b, n);
    source = src;
  }
}
