/* 
 *  $Id: CodeGroup.Java,v 1.9 1998/05/29 20:51:47 gback Exp $
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */


package ANTS_PKG;
import NODEOS_PKG.*;

import java.util.*;
import UTILS_PKG.*;

/**
 * Record collecting information about methods into a related group.
 * Stored in the cache while code is resident or being loaded.
 *
 * @author David Wetherall
 */

class CodeGroup {
  Vector codeIDs;
  Vector meths;
  TypeID groupID;
  Vector groupIDs;
  TypeID protocolID;
  int parts;
  int chunks;
  CodeGroup next;
  CodeGroup prev;
  boolean pin;
  int builtin;
  FlowHandle flowHandle;

  private TypeID computeGroupID() {
    TypeID.sortVector(codeIDs);
    MD5 id = new MD5();
    for (int i = 0; i < codeIDs.size(); i++) {
      TypeID t = (TypeID)(codeIDs.elementAt(i));
      id.update(t.unwrap());
    }
    return TypeID.wrap(id.hash());
  }
 
  void setGroupID() {
    groupID = computeGroupID().intern();
  }
  
  boolean checkGroupID() {
    groupID = groupID.intern();
    return groupID.equals(computeGroupID());
  }

  private TypeID computeProtocolID() {
    TypeID.sortVector(groupIDs);
    MD5 id = new MD5();
    for (int i = 0; i < groupIDs.size(); i++) {
      TypeID t = (TypeID)(groupIDs.elementAt(i));
      id.update(t.unwrap());
    }
    return TypeID.wrap(id.hash());
  }
  
  void setProtocolID() {
    protocolID = computeProtocolID().intern();
  }
  
  void setChunks() throws Exception {
    for (int i = 0; i < meths.size(); i++) {
      Method m = (Method)(meths.elementAt(i));
      chunks += pieces(m.code);
    }

    // start with all parts present
    parts = -1;

    // sanity check limit
    if (chunks > 32)
      throw new Exception("CodeGroup "+ groupID +" too long");
  }

  final static int CHUNK = 1024;

  final short part(int i) {
    return (short)((1<<i) | (-1)<<chunks);
  }

  static int pieces(byte[] code) {
    return (code.length+CHUNK-1) / CHUNK;
  }

  final void enqueue(Method meth, Buffer pkt) {
    if (flowHandle.getQueueLength() < Method.MAXWAITERS) {
      pkt.setObject(meth);
      flowHandle.addBufferToReceiveQueue(pkt);
    }
  }

  final void doactivate(Node n, FlowHandle flowHandle) {
    // rest of activation, done in the context of new channel/process
    for (int i = 0; i < meths.size(); i++) {
      Method m = (Method)meths.elementAt(i);

      // verify code, group and protocol match method
      if (!m.checkMethodID()) {
	Entity.error("Bad protocol identifier");
      }

      // bootstrap
      m.registerIDs();

      // if I'm asked to use a different flow handle, merge my packets in
      if (this.flowHandle != flowHandle)
	flowHandle.merge(this.flowHandle);

      // make sure this method is associated with the protocol's flow
      m.setFlowHandle(flowHandle);
    }
  }

  final void activate(Node n) {
    // this part is done in the context of DLResponseCapsule (for now at least)
    for (int i = 0; i < meths.size(); i++) {
      Method m = (Method)meths.elementAt(i);
      m.setCodeID();
      codeIDs.addElement(m.codeID);
    }

    if (!checkGroupID())
      Entity.error("Bad code group identifier");

    setProtocolID();    

    // see if this protocol is already running
    InChannel inChannel = n.findInChannel(protocolID);

    if (inChannel == null) {
	// now start up channel, waiting capsules have already piled up
	// and will be processed automatically

	// note that we don't provide a flow handle to this InChannel quite yet -
	// the channel will grab or create one during its initialization
	inChannel = new InChannel(n, true /* replenish */);

	// attach it right away
	n.attachInChannel(protocolID, inChannel);

	// give that guy some leeway...
	flowHandle.addBuffers(25, ((UDPChannel)n.link).BUFSIZE);

	// tell channel to activate this code group before processing packets
	inChannel.addCodeGroup(this);
	inChannel.start();
    } else {
	// else just add this group, the channel will initialize it later
	inChannel.addCodeGroup(this);
    }
  }

  final void add(Method m) {
    m.group = this;
    meths.addElement(m);
    // merge all the buffers piled up so far into our flow
    flowHandle.merge(m.flowHandle);
    // concatenate names
    flowHandle.setName(m.flowHandle.getName() + "/" + flowHandle.getName());
  }

  CodeGroup() {
    this.meths = new Vector();
    this.codeIDs = new Vector();
    this.groupIDs = new Vector();
    this.builtin = 0;
    flowHandle = new EmptyFlowHandle("");
  }

  CodeGroup(TypeID groupID) {
    this();
    this.groupID = groupID;
    this.builtin = 0;
  }
}


