/* 
 *  $Id: ConfigurationManager.Java,v 1.7 1998/06/01 05:46:06 gback Exp $
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */


package ANTS_PKG;

import java.awt.event.*;
import java.awt.*;

import java.io.*;
import java.net.*;
import java.util.*;

import UTILS_PKG.*;


/**
 * manager that creates node/channel/app configurations
 * as specified in a control file
 *
 * @author David Wetherall
 */


public class ConfigurationManager
  extends Manager implements Runnable, ItemListener
{
  private boolean gui;
  // after startup, should I continue to accept commands via a 
  // telnet connection? If nonzero, that's the port to be used.
  private int consoleport;
  private Checkbox[][] cbs;
  private Frame window;

  final static String[] defaults = { "-gui", " false", 
		"-consoleport", "0" 	// console disabled by default
	    };
  
  public void setArgs(KeyArgs k) 
    throws Exception
  {
    k.merge(defaults);

    for (int i = 0; i < k.length(); i++) {
      if (k.key(i).equals("-gui")) {
	gui = (new Boolean(k.arg(i))).booleanValue();
	k.strike(i);
      }
      if (k.key(i).equals("-consoleport")) {
	consoleport = (new Integer(k.arg(i))).intValue();
	k.strike(i);
      }
    }
    
    super.setArgs(k);
  }

  public void itemStateChanged(ItemEvent evt) {
    int c = 0;
    for (EntityList e = children; e != null; e = e.next) {
      for (int i = 0; i < LOG_LEVELS; i++)
	if (evt.getItemSelectable() == cbs[c][i]) {
	  if (evt.getStateChange() == ItemEvent.SELECTED)
	    e.entity.setLogLevel(e.entity.getLogLevel() | L[i]);
	  else 
	    e.entity.setLogLevel(e.entity.getLogLevel() & ~L[i]);
	}
      c++;
    }
  }
 
  void createPanel() {
    Panel n = new Panel();
    n.setLayout(new GridLayout(0, 1));
    Panel p = new Panel();
    p.setLayout(new GridLayout(0, LOG_LEVELS));
    cbs = new Checkbox[children.length()][LOG_LEVELS];
    int c = 0;
    for (EntityList e = children; e != null; e = e.next) {
	n.add(new Label(e.entity.getName() + ": "));
	for (int i = 0; i < LOG_LEVELS; i++) {
	    boolean on = (e.entity.getLogLevel() & L[i]) != 0;
	    cbs[c][i] = new Checkbox(Integer.toString(i), null, on);
	    cbs[c][i].addItemListener(this);
	    p.add(cbs[c][i]);
	}
	c++;
    }
      
    window.add("East", n);
    window.add("West", p);
    window.pack();
    window.show();
  }

  public synchronized void refreshLoggingPanel()
  {
    if (!gui) return;

    if (LOGGING) {
	window.removeAll();
	window.dispose();
	createPanel();
    }
  }

  public void start() {
    // createPanel now done in refreshLoggingPanel
  }

  Socket              client; 
  PrintWriter         out;
  LineNumberReader    in;
  Node n = null;

  public ConfigurationManager(Node n, Socket client)
    throws Exception
  {
    this.n = n;
    this.client = client;
    out = new PrintWriter(client.getOutputStream());
    in = new LineNumberReader(new InputStreamReader(client.getInputStream()));
  }

  /**
   * handle commands that are identical for remote console and config file
   */
  private String handleLine(String []args, boolean start) 
    throws Exception 
  {
    if (args[0].equals("application")) {
      if (args.length < 3) throw new Exception("short application");
      Class appClass = Class.forName(args[2]);
      Application a = (Application)(appClass.newInstance());    
      a.setBoss(this);
      a.setName(args[2]);
      a.setArgs(args, 3);
      a.attachNode(n);
      if (start)
	  a.start();
      return "Starting application " + a.getName();
    }
    return "Sorry, command not understood.\n" 
      + "Try `who', `quit', `shutdown', or `application ...'";
  }

  /**
   * Read-and-Execute Loop
   */
  public void run()
  {
    out.println("Welcome at ANTS Node " + n.getName());
    out.println("Type `quit' to quit.");
    out.flush();

    try {
      for (;;) {
	LineListTokenizer lt = new LineListTokenizer(
	    new StringReader(in.readLine() + "\n"));

        String[] args = lt.list();
	if (args == null)
	    continue;

        if (args[0].equals("shutdown")) {
	  System.exit(-1);
        } else if (args[0].equals("quit")) {
	  throw new Exception("done!");
	} else if (args[0].equals("who")) {
	  // what's going on on this node?
	  out.println("Currently registered protocols:");
	  for(Enumeration e = n.protIDToChannel.keys(); e.hasMoreElements(); ) {
	    TypeID protID = (TypeID)e.nextElement();
	    InChannel in = n.findInChannel(protID);
	    out.println(protID + "\t" + in.getName());	// XXX more info!
	  }
	} else {
	  // handle command and send reply
   	  out.println(handleLine(args, true));
	}
	out.flush();
      }
    } catch (Exception e) {
	;
    } finally {
      out.println("exiting...");
      out.flush();
      try {
        client.close();
      } catch (Exception e) {}
    }
  }

  public ConfigurationManager(String[] params)
    throws Exception
  {
    super();
    
    // two required args (ignore rest of command-line args for now)

    String file = params[0];
    String node = params[1];

    setName("Manager@" + params[1]);
    KeyArgs k = new KeyArgs(params, 2);

    // process the configuration file, creating entities

    LineListTokenizer lt = new LineListTokenizer(file);

    String[] args;
    while ((args = lt.list()) != null) {

      // note, we ignore commands that we don't know about

      if (args.length < 2) continue;
      if (!node.equals(args[1])) continue;

      if (args[0].equals("node")) {
	n = new Node(this, args[1], args[1]);
	n.setArgs(args, 2);

      } else if (args[0].equals("channel")) {
	if (args.length < 3) throw new Exception("short channel");
	Channel c = Channel.New(this, args[2], args[2]);
	c.setArgs(args, 3);
	c.attachNode(n);

      } else if (args[0].equals("extension")) {
	if (args.length < 3) throw new Exception("short extension");
	Class extClass = Class.forName(args[2]);
	Extension e = (Extension)(extClass.newInstance());    
	e.setBoss(this);
	e.setName(args[2]);
	e.setArgs(args, 3);
	e.attachNode(n);

      } else if (args[0].equals("manager")) {
	k.merge(args, 2);
	setArgs(k);

      } else 
	handleLine(args, false);
      
    }

    // start ourselves, then rely on the node
    // to start others in a safe order
    start();
    n.start();

    if (gui) {
      window = FrameCreator.createFrame(getName());
      createPanel();
    }

    if (consoleport > 0) {
	try {
            ServerSocket        s = new ServerSocket(consoleport);
            for (;;)
                new Thread(new ConfigurationManager(n, s.accept())).start();
        } catch (Exception e) {
            e.printStackTrace(System.out);              
        }
    }
  }
  
  public static void main(String[] args)
    throws Exception
  {
    String use = "<file> <node> [keyargs]";
    String cl = "ants.ConfigurationManager";
    if (args.length < 2) {
      System.out.println("usage: " + cl + " " + use);
      System.exit(1);
    }
    
    new ConfigurationManager(args);
  }
  
}
