/* 
 *  $Id: DLClassLoader.Java,v 1.6 1998/05/29 06:00:45 gback Exp $
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */


package ANTS_PKG;

import java.util.Hashtable;


/**
 * Demand Loading ClassLoader.  Presents an interface to load a
 * code group from byte arrays received by the DL protocol or 
 * registered at the Node.
 *
 * @author David Wetherall
 */


// warning -- there be wicked interactions here.  this code is 
// heavily commented and asserted as a result. see debug also.

final class DLClassLoader extends ClassLoader {

  /**
   * Cache of classes loaded by all local DLClassLoaders
   * into this java vm.  Used to avoid multiple definitions.
   */

  private static Hashtable sharedRegistry = new Hashtable();

  /**
   * Cache of classes loaded by this DLClassLoader. Used to
   * determine what is accessibile in the capsule loading space.
   */

  private Hashtable registry;

  /** Side information for loadClass -- the code group being loaded */
  private CodeGroup inProgress;
  
  /** Compile time flag for debugging statements */
  private static boolean DEBUG = false;

  /** Debug print statement, redirects to logger */
  private static void debug(String msg) {
    if (DEBUG) System.err.println("DLClassLoader: "+ msg);
  }

  synchronized Class lookupClassInRegistry(String name) 
  {
    Class cl = (Class)registry.get(name);
    if (cl != null) {
      debug("loadClass: result found in registry for "+ name);
    }

    // fall back on shared registry
    cl = (Class)sharedRegistry.get(name);
    if (cl != null) {
      debug("loadClass: result found in shared registry for "+ name);
    }
    return cl;
  }

  /**
   * This method is the catch-all to intercept any unknown class 
   * references that are needed by capsules.  Eventually all classes
   * should be found already in the registry, or an error occurs.
   */

  protected synchronized Class loadClass(String name, boolean resolve) 
    throws ClassNotFoundException
  {
    CodeGroup g = inProgress;
    Class cl = lookupClassInRegistry(name);

    // this is a group load, not already in the registry,
    // so it must be another class in the code group ...

    for (int i = 0; i < g.meths.size(); i++) {
	Method m = (Method)g.meths.elementAt(i);

	// first, we find the right method
	if (name.equals(m._cn)) {

	  debug("loadClass: calling defineClass for "+ name);
	  cl = defineClass(name, m.code, 0, m.code.length);

	  // about the only thing we can do is check afterwards
	  // in the hope of catching prior definitions

	  Entity.assert(cl.getClassLoader() == this,
			  "DLClassLoader: multiple definitions for "+ name);

	  // we've tried all cases, done
	  break;
      }
    }

    // if we still don't have it, then breach the load boundary
    // this is for debugging purposes only
    if (cl == null) {
      debug("breach load for "+ name);
      Entity.warn("DLClassLoader: breach load for "+ name);
      cl = findSystemClass(name);
    }
    
    // be more graceful here?
    Entity.assert(cl != null, "DLClassLoader: empty load for "+ name);
    Entity.assert(cl.getName().equals(name),
		  "DLClassLoader: unexpected name ("+ 
		  cl.getName() +") for "+ name);
    
    // update registry with the right info
    if (g.builtin != 0) 
      sharedRegistry.put(name, cl);
    else
      registry.put(name, cl);
    
    if (resolve) {
      debug("loadClass: calling resolveClass for "+ name);
      resolveClass(cl);
    }
    
    debug("loadClass: returning for "+ name);
    return cl;
  }


  /**
   * Entry point used by a node to register code groups.
   * Loads ultimately come from the local filesystem
   */

  synchronized static boolean loadForRegister(CodeGroup g, boolean share) 
    throws Exception
  {
    debug("loadForRegister: begin loading "+ g);

    try {
      
      for (int i = 0; i < g.meths.size(); i++) {
	Method m = (Method)g.meths.elementAt(i);
	
	// load failures will throw an exception here, which
	// should percolate up to the application registering
	debug("loadForRegister: finding via class.forname for "+ m._cn);
	m.cl = Class.forName(m._cn);
	if (share)
	    sharedRegistry.put(m._cn, m.cl);
      }
    } catch (Exception e) {
      throw e;
    }

    debug("loadForRegister: end loading "+ g);
  
    return true;
  }


  /**
   * Entry point used by a the DL protocol load code groups.
   * Loads ultimately come from byte arrays.
   */

  synchronized boolean loadOnDemand(CodeGroup g) {
    debug("loadOnDemand: begin loading "+ g);

    boolean loaded = true;
    {
      inProgress = g;
      Class	[]cl = new Class[g.meths.size()];
      
      for (int i = 0; i < g.meths.size(); i++) {
	Method m = (Method)g.meths.elementAt(i);
	
	debug("loadOnDemand: trial loading "+ m._cn);
	
	try {
	  cl[i] = loadClass(m._cn, true);
	} catch (ClassNotFoundException e) {
	  loaded = false;
	  break;
	}
      }
      
      if (loaded) {
	for (int i = 0; i < g.meths.size(); i++)
	  ((Method)g.meths.elementAt(i)).cl = cl[i];
      }
    } 
    
    debug("loadOnDemand: end loading "+ g);
    return loaded;
  }
  

  /**
   * Export node and extension classes that capsules may access
   * by adding them to the registry.
   */

  public static synchronized void exportClass(String name) 
    throws ClassNotFoundException
  {

    debug("exportClass: adding to registry for "+ name);
 
    // all exported classes must be locally loadable
    Class cl = Class.forName(name);
    sharedRegistry.put(name, cl);
  }
  
  /** Convenience callback for capsules to check their load path */
  public void verifyOrigin(String name) {
    if (registry.get(name) == null)
      Entity.error("DLClassLoader: "+ name +" fails load origin check");
  }
  
  /** One DLClassLoader is constructed for each Protocol */
  DLClassLoader() {
    registry = new Hashtable();
  }
}



