/* 
 *  $Id: DLResponseCapsule.Java,v 1.6 1998/05/29 06:00:46 gback Exp $
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */


package ANTS_PKG;

/**
 * Demand Load Response capsule and processing
 *
 * @author David Wetherall
 */


public class DLResponseCapsule extends Capsule {

  // WARNING: this is a builtin capsule, whose version number
  // must be manually kept in synch with  its functionality
  final public static String VERSION = "ants.DLResponseCapsule 10/22/97";

  final private static byte[] MID = findMID("ants.DLResponseCapsule");
  protected byte[] mid() { return MID; }

  final private static byte[] PID = findPID("ants.DLResponseCapsule");
  protected byte[] pid() { return PID; }

  private TypeID target;
  private TypeID group;
  private TypeID[] groups;
  private int part;
  private int clen;
  private int coff;
  private String _cn;
  private ByteArray code;

  public int length() {
    int s = Xdr.TYPEID + Xdr.TYPEID + Xdr.SHORT + groups.length*Xdr.TYPEID +
      Xdr.INT + Xdr.INT + Xdr.INT + Xdr.STRING(_cn) + Xdr.BYTEARRAY(code);
    return super.length() + s;
  }
  
  public Xdr encode() {
    Xdr xdr = super.encode();
    xdr.PUT(target);
    xdr.PUT(group);
    xdr.PUT((short)groups.length);
    for (int n = 0; n < groups.length; n++)
      xdr.PUT(groups[n]);
    xdr.PUT(part);
    xdr.PUT(clen);
    xdr.PUT(coff);
    xdr.PUT(_cn);
    xdr.PUT(code);

    return xdr;
  }

  public Xdr decode() {
    Xdr xdr = super.decode();
    target = xdr.TYPEID();
    group = xdr.TYPEID();
    int l = xdr.SHORT();
    groups = new TypeID[l];
    for (int n = 0; n < l; n++) {
      groups[n] = xdr.TYPEID();
    }
    part = xdr.INT();
    clen = xdr.INT();
    coff = xdr.INT();
    _cn = xdr.STRING();
    code = xdr.BYTEARRAY();
    
    return xdr;
  }
  
  public boolean evaluate(Node n) {
    if (Entity.LOGGING)
      n.log(Entity.L[0], "decode: received DLResponse "+
	    target +" from "+ NodeAddress.toString(getSrc()));
    
    Method m = n.capsuleCode.lookup(target);
    if (m == null) {
      // this would be an unexpected DLResponseCapsule - otherwise 
      // DLBootstrap would make sure it's in the code cache
      m = new Method(target);
      n.capsuleCode.enter(m);
    }

    // mark as answer forthcoming, so repeat requests are allowed
    m.repeat = 0;
    m.flowHandle.setName(_cn);

    CodeGroup g = n.capsuleCode.retrieve(group);
    if (g == null) {
      g = new CodeGroup(group);
      n.capsuleCode.set(g);
    } 

    if (m.group == null) g.add(m);

    if (g.groupIDs.size() ==  0) {
      for (int i = 0; i < groups.length; i++) {
	g.groupIDs.addElement(groups[i]);
      }
    }

    if ((g.parts | part) != g.parts) {
      if (m.code == null) {
	
	if (Entity.LOGGING)
	  n.log(Entity.L[0], "DLResponse: allocating "+ clen +" for "+ _cn);
	
	m.code = new byte[clen];
	m._cn = _cn;
	g.chunks += g.pieces(m.code);
      }
      
      if (Entity.LOGGING)
	n.log(Entity.L[0], "DLResponse: adding to "+ _cn + " (off= "+
	      coff +", part= "+ Integer.toHexString(part) + ", len= " +
	      code.length() + ")");
      
      code.getBytes(0, m.code, coff, code.length());
      g.parts |= part;
      
      if (g.parts == -1) {
	
	if (Entity.LOGGING)
	  n.log(Entity.L[0], "DLResponse: completed with "+ _cn);

	// activate new code group
	g.activate(n);

      }
    } else if (Entity.LOGGING)
      n.log(Entity.L[0], "decode: unneeded DLResponse "+
	    target +" from "+ NodeAddress.toString(getSrc()));
    
    return true;
  }
  

  DLResponseCapsule() { }

  DLResponseCapsule(Method m, int idx, int pos, int len) {
    target = m.methodID;
    group = m.group.groupID;
    groups = new TypeID[m.group.groupIDs.size()];
    for (int i = 0; i < m.group.groupIDs.size(); i++) {
      groups[i] = (TypeID)(m.group.groupIDs.elementAt(i));
    }
    part = m.group.part(idx);
    clen = m.code.length;
    coff = pos;
    _cn = m._cn;

    code = new ByteArray(m.code, coff, len);
  }
}


