/* 
 *  $Id: Method.Java,v 1.8 1998/05/29 20:51:49 gback Exp $
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */

package ANTS_PKG;
import NODEOS_PKG.*;

import java.util.*;
import UTILS_PKG.*;


/** 
 * This class (along with CodeGroups) is used to represent all the
 * information about a type of capsule in the code cache.  It is used
 * only within ants, from Protocol during registration, Channel during
 * capsule execution dispatch, and DL* during code loading.
 *
 * @author David Wetherall
 */

final class Method {

  // the parts of method have package access since it may be in many
  // different states of completion during code loading and protocol
  // registration

  byte[] code;
  String _cn;
  Class cl;
  TypeID codeID;
  int hash;
  TypeID methodID;
  CodeGroup group;
  FlowHandle flowHandle;
  Method next;
  int repeat;

  // bytecodereader that uses default class path 
  private static ByteCodeReader	classCode = new ClassPathReader();

  final static int MAXWAITERS = 4;

  // until we have a better way to initialize the 
  // methodid of a capsule, we keep this side table
  // based on local names
  
  /** Entry point used for protocol definition and loading */
  void registerIDs() {
    TypeID[] ids = new TypeID[3];
    ids[0] = methodID;
    ids[1] = group.groupID;
    ids[2] = group.protocolID;

    InChannel.registerIDs(_cn, ids);
  }

  /** Access point to initialize capsule methodIDs */
  static byte[] findMID(String name) {
    return InChannel.currentInChannel().findMID(name);
  }
  
  /** Access point to initialze capsule groupIDs */
  static byte[] findGID(String name) {
    return InChannel.currentInChannel().findGID(name);
  }
  
  /** Access point to initialze capsule protocolIDs */
  static byte[] findPID(String name) {
    return InChannel.currentInChannel().findPID(name);
  }
  
  private TypeID computeMethodID() {
    MD5 id = new MD5();

    // code, group, protocol is the canonical order
    id.update(codeID.unwrap());
    id.update(group.groupID.unwrap());
    id.update(group.protocolID.unwrap());
    return TypeID.wrap(id.hash());
  }

  void setMethodID() {
    methodID = computeMethodID().intern();
  }

  void setCodeID() { 
    codeID = TypeID.wrap(MD5.hash(code));
  }

  boolean checkMethodID() {
    methodID = methodID.intern();
    return methodID.equals(computeMethodID());
  }

 /**
  * Dispatch a buffer.
  * Simply hand it to the associated flow handle
  */
  public Buffer dispatchBuffer(Buffer b, int offset) {
    b.setObject(this);
    return flowHandle.dispatchBuffer(b, offset);
  }
   
  // store packets in FIFO order, keeping only the last MAXWAITERS 
  void enqueue(Buffer pkt) {
    if (flowHandle.getQueueLength() < MAXWAITERS) {
      pkt.setObject(this);
      flowHandle.addBufferToReceiveQueue(pkt);
    }
  }

  /** set the flow handle associated with that method */
  void setFlowHandle(FlowHandle flowHandle)
  {
    this.flowHandle = flowHandle;
  }

  /**
   * A new method is created, either as the result of receiving an unknown
   * capsule or as the result of receiving code for a method while processing
   * a DLResponse capsule.
   */
  Method(TypeID methodID) {
    this.flowHandle = new EmptyFlowHandle("");
    // no buffer reservation for free queue needed
    this.methodID = methodID;
    this.repeat = 0;
  }

  /** Construct and initialize from the local class name */
  Method(String name) throws Exception {

    // until we control the local name under which 
    // a method is loaded, we need to save this info
    _cn = name;
    
    code = classCode.readByteCode(name);
    setCodeID();
    repeat = 0;
  }
}
