/* 
 * Copyright (c) 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * Permission to use, copy, modify and distribute this software is hereby
 * granted provided that (1) source code retains these copyright, permission,
 * and disclaimer notices, and (2) redistributions including binaries reproduce
 * the notices in supporting documentation, and (3) neither the name of the
 * University nor the Flux Group may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 * 
 * THE UNIVERSITY OF UTAH AND THE FLUX GROUP ALLOW FREE USE OF THIS SOFTWARE
 * IN ITS "AS IS" CONDITION, AND DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY
 * DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * The University requests users to return to flux-dist@cs.utah.edu any
 * improvements that they make and grant the University redistribution rights.
 */

package NODEOS_PKG;

/**
 * A dispatcher for ANEP version 1
 * <p>
 * This is a dispatcher for the <b>Active Network Encapsulation Protocol</b>
 * as specified in the RFC draft by the Active Networks Group. 
 * <p>
 * For more information, consult the
 * <a href = "http://www.cis.upenn.edu/~switchware/ANEP">ANEP homepage</a>
 * <p>
 * For more information on AN type ids, see the
 * <a href = "http://www.cs.utah.edu/~gback/fast/res/an-assigned-numbers.txt">
 *	list of assigned numbers.</a>
 *
 * @author 	Godmar Back
 */
public class ANEPDispatcher implements FlowDispatcher
{
    public  static final short ANTS_ID = 18;
    public  static final short PLAN_ID = 19;
    public  static final short SPANNER_ID = 24;

    private static final byte ANEP_VERSION = 1;
    private static final int ANEP_VERSION_OFFSET = 0;
    private static final int FLAG_OFFSET = 1;
    private static final int TYPEID_OFFSET = 2;
    private static final int HEADERLEN_OFFSET = 4;
    private static final int PACKETLEN_OFFSET = 6;

    /* currently supported maximum ANid */
    private static final int  MAX_ANID = 256;	

    private FlowDispatcher	anFlows[];
    private FlowDispatcher	defaultFlow;

    /**
     * Class constructor.
     */
    public ANEPDispatcher()
    {
	anFlows = new FlowDispatcher[MAX_ANID];
    }

    /**
     * Dispatch a given buffer.
     *
     * This will look at the AN id in the buffer. If a flow dispatcher has been
     * registered for that id, the buffer will be dispatched to that flow.
     * <p>
     * If an AN id is not recognized and the discard bit is not set, it will
     * dispatch the buffer to a default flow, if one is registered.
     * <p>
     * Note that only one flow dispatcher per id is supported.
     * <p>
     * See <a href="flow.FlowDispatcher.html#dispatchBuffer">FlowDispatcher</a>
     * for a description of the parameters.
     *
     * @see Buffer
     * @see FlowDispatcher#dispatchBuffer
     */
    public Buffer dispatchBuffer(Buffer b, int offset)
    {
	byte []data = b.getData();
	boolean discard = (data[offset + FLAG_OFFSET] & 0x80) != 0;

	int typeid = EXTRACT_SHORT(data, offset + TYPEID_OFFSET);

	/* header length is in 32 bit words */
	int hlen = EXTRACT_SHORT(data, offset + HEADERLEN_OFFSET) << 2;

	// DEBUG("dispatch ANEP typeid = " + typeid);
	if (data[offset] == ANEP_VERSION) {
	    if (anFlows[typeid] != null)
		return anFlows[typeid].dispatchBuffer(b, offset + hlen);

	    if (defaultFlow != null && !discard)
		return defaultFlow.dispatchBuffer(b, offset);
	}
	return null;
    }

    /**
     * Register a flow dispatcher under a certain id.
     *
     * @param id	a integer containing the AN id
     * @param fh	the flow dispatcher to be associated with the id
     */
    public void registerANid(int id, FlowDispatcher fh)
    {
	anFlows[id] = fh;
    }

    /**
     * Register a flow to which unrecognized packets are to be dispatched.
     *
     * @param	fh	a flow dispatcher to which unrecognized packets go.
     */
    public void registerDefault(FlowDispatcher fh)
    {
	defaultFlow = fh;
    }
}
