/* 
 * Copyright (c) 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * Permission to use, copy, modify and distribute this software is hereby
 * granted provided that (1) source code retains these copyright, permission,
 * and disclaimer notices, and (2) redistributions including binaries reproduce
 * the notices in supporting documentation, and (3) neither the name of the
 * University nor the Flux Group may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 * 
 * THE UNIVERSITY OF UTAH AND THE FLUX GROUP ALLOW FREE USE OF THIS SOFTWARE
 * IN ITS "AS IS" CONDITION, AND DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY
 * DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * The University requests users to return to flux-dist@cs.utah.edu any
 * improvements that they make and grant the University redistribution rights.
 */

package NODEOS_PKG;

/**
 * A class for buffers.
 * <p>
 * Note: There will be an interface or an abstract base class describing
 * common properties of buffers.
 *
 * (We would prefer to use an interface here, but Pat pointed
 * out that the dispatch penalty in Kaffe is very high.)
 * <p>
 * For now, a buffer is represented by an array of bytes.
 * It also keeps track of a start index within the buffer, called offset.
 * <p>
 * Buffers can be dependent on other buffers or have dependents itself.
 * These buffers form a singly linked list.  The first buffer in the list
 * contains the actual data.  A buffer can be made independent, which removes
 * it from that list and copies the data, if necessary.
 * <p>
 * The purpose of dependent buffers is to avoid actually copying the data
 * if a buffer must be dispatched to multiple destinations.  Instead, the
 * buffer is marked as dependent - which can be done in constant time. 
 * Making it independent, which involves copying the data in the buffer,
 * is done in the application's context.
 *
 * @author Godmar Back
 */

public final class Buffer extends Linkable {

    byte[] buf;
    int offset;
    int len;

   /**
    * Return current size of the object.
    * The size is the number of bytes after the start offset.
    *
    * @return current length of the buffer
    */ 
    public final int getLength() 
    { 
	return len; 
    }

   /**
    * Class constructor - allocates a buffer of a given size.
    * 
    * @param	len	an int specifying the size of the buffer
    */
    public Buffer(int len)
    {
	buf = new byte[len];
	offset = 0;
	this.len = len;
    }

   /**
    * Class constructor - makes a buffer out of an array of bytes.
    *
    * @param buf	array of bytes used by this buffer.
    */
    public Buffer(byte[] buf)
    {
	this.buf = buf;
	this.offset = 0;
	this.len = buf.length;
    }

   /**
    * Class constructor - makes a buffer out of an array of bytes with a
    *	given initial offset.
    *
    * @param buf		array of bytes used by this buffer.
    * @param offset	initial offset
    */
    public Buffer(byte[] buf, int offset)
    {
	this(buf, offset, buf.length-offset);
    }

    public Buffer(byte[] buf, int offset, int len) 
       throws IndexOutOfBoundsException
    {
	if (offset + len > buf.length || offset < 0 || len < 0)
	    throw new IndexOutOfBoundsException();
	this.buf = buf;
	this.offset = offset;
	this.len = len;
    }

    /**
     * beginning of my additions for dependent/independent buffers
     */

    /**
     * Flag saying whether buffer was already exchanged in the
     * dispatch process.  Upon dispatch, this flag is set to false.
     * The first flow handle which will enqueue the buffer will
     * exchange it with a free buffer from its queue and set the
     * flag to true, causing subsequent handles to make their buffer
     * dependent on this one.
     */
    boolean	exchanged;

    /**
     * Print human-readable statistics about this buffer.
     */
    public String toString()
    {
	String name = super.toString() 
		+ " off=" + offset 
		+ " len=" + len 
		+ " cap=" + buf.length;

	if (master != null)
	    name += " C";

	if (next_copy != null)
	    if (master == null)
		name += " M";
	    else
		name += " NC";
	return name;
    }

    /**
     * Resets buffer to initial capacity.
     *
     * This resets the offset to zero, restoring its original length.
     */
    public Buffer reset()
    {
	offset = 0;
	len = buf.length;
	return this;
    }

    /**
     * Advance buffer by a given offset.
     *
     * @param	off	offset by which to increment start.
     */
    void advanceStartBy(int off)
    {
	offset += off;
	len -= off;
    }

    /** 
     * Read an amount of bytes.
     * <p> 
     * Read an amount of bytes from offset into buf.  This routine takes
     * dependent buffers into account.
     *
     * @param buf	destination buffer
     * @param offset	offset in destination buffer
     * @param amount	number of bytes to be read.
     * @return actual number of bytes that were read.
     */
    public int read(byte []buf, int offset, int amount)
    {
	/* if the data is still in the master's buffer, return it from there */
	boolean amCopy = master != null;
	byte []data = amCopy ? master.buf : this.buf;
	long length = amCopy ? master.len : this.len;

	int maxamount = (int)(len - offset);
	if (amount > maxamount)
	    amount = maxamount;
	System.arraycopy(data, (int)(this.offset + offset), buf, 0, amount);
	return amount;
    }

    /** 
     * Write an amount of bytes.
     * <p> 
     * Write an amount of bytes to offset into buf.  This routine does not
     * take dependent buffers into account. That is, it might overwrite data
     * in a master, and the data it writes to a dependent buffer might be
     * overwritten when that buffer is made independent. (XXX)
     *
     * @param buf	destination buffer
     * @param offset	offset in destination buffer
     * @param amount	number of bytes to be read.
     * @return actual number of bytes that were read.
     */
    public int write(byte []buf, int offset, int amount)
    {
	int maxamount = (int)(len - offset);
	if (amount > maxamount)
	    amount = maxamount;
	System.arraycopy(buf, 0, this.buf, (int)(this.offset + offset), amount);
	return amount;
    }

    /**
     * Sets  current size of the object.
     * The size is the number of bytes after the offset.
     */ 
    public void setLength(int len)
    {
	if (offset + len <= buf.length)
	    this.len = len;
    }

    /**
     * Return the underlying byte[] array.
     * This returns the byte array upon which the buffer is formed.
     * Note that Java doesn't have pointer arithmetic; use getStart.
     *
     * @see #getStart
     */
    public byte[] getData()
    {
	return buf;
    }

    /**
     * Return the start offset of the buffer
     *
     * @return start offset of first valid byte in buffer
     * @see #getData
     */
    public long getStart()
    {
	return offset;
    }

    /*------------------- methods dealing with shadowing -------------------*/

    /* the buffer containing the actual data; null if data is in our buffer */
    private Buffer	master;

    /* next guy in chain of copies */
    private Buffer	next_copy;

    /**
     * Check whether this buffer is independent.
     *
     * @return true if this buffer is not dependent on any other buffer and
     * 	if no other buffer is dependent on us.
     */
    boolean isIndependent() 
    {
	return master == null && next_copy == null;
    }

    /**
     * Make this buffer a dependent of another buffer.
     *
     * @param original buffer upon which this buffer will depend.
     */
    void makeDependentOf(Buffer original)
    {
	if (original.master != null)
	    master = original.master;
	else
	    master = original;

	/* insert us in our master's list of copies */ 
	this.next_copy = master.next_copy;
	master.next_copy = this;
    }

    /**
     * Make this buffer independent of any other buffer.
     * <p>
     * This could be called on-demand on a map or write.
     */
    void makeIndependent()
    {
	/* we have the data */
	if (master == null) {	
	    /* are we somebody else's master? */
	    if (next_copy != null) { 
		synchronized (this) {
		    int len = buf.length;
		    if (len > next_copy.buf.length)
			len = next_copy.buf.length;
		    System.arraycopy(this.buf, 0, next_copy.buf, 0, len);
		    /* now declare `next_copy' the master for the rest 
		       of the chain 
		     */
		    for (Buffer b = next_copy; b != null; b = b.next_copy)
			b.master = next_copy;
		    next_copy.master = null;
		    next_copy = null;
		}
	    }
	} else {
	    synchronized (master) {
		/* our master has the data, we must copy them */
		int len = master.buf.length;
		if (len > buf.length)
		    len = buf.length;
		System.arraycopy(master.buf, 0, this.buf, 0, len);

		/* now remove us from the list of copies */
		for (Buffer b = master; b.next_copy != null; b = b.next_copy) {
		    if (b.next_copy == this) {
			b.next_copy = next_copy;
			break;
		    }
		}
		master = next_copy = null;
	    }
	}
    }

    // a placeholder for an Object that can be associated with that buffer
    private Object	obj;

    /**
     * associate an Object with this buffer
     */
    public void setObject(Object obj)
    {
	this.obj = obj;
    }

    /**
     * retrieve the object associated with this buffer
     */
    public Object getObject()
    {
	return obj;
    }

    /* missing: ? getblocksize, setsize, unmap, wire, unwire, copy */
}

