/* WARNING: This file was automatically generated from nodeos/FlowHandle.Java DO NOT EDIT *//* 
 * Copyright (c) 1998 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * Permission to use, copy, modify and distribute this software is hereby
 * granted provided that (1) source code retains these copyright, permission,
 * and disclaimer notices, and (2) redistributions including binaries reproduce
 * the notices in supporting documentation, and (3) neither the name of the
 * University nor the Flux Group may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 * 
 * THE UNIVERSITY OF UTAH AND THE FLUX GROUP ALLOW FREE USE OF THIS SOFTWARE
 * IN ITS "AS IS" CONDITION, AND DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY
 * DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * The University requests users to return to flux-dist@cs.utah.edu any
 * improvements that they make and grant the University redistribution rights.
 */

package nodeos ;

import java.net.*;
import java.io.IOException;

/**
 * A handle on a flow of packets.
 * <p>
 * Flow handles serve three purposes, each of which is optional.
 * <ul>
 * <li>Packet buffering.  Every flow is associated with a queue of free
 *     buffers and a queue of already received, but undelivered packets.
 *     Packets cannot be buffered unless free buffers are available.
 *     It is the responsibility of the user of the handle to provide
 *     it with buffers beforehand.
 * <p>
 * <li>Payload dispatch. A flow handle may be associated with a dispatcher.
 *     In this case, the payload of all packets dispatched to the handle 
 *     will be dispatched to that dispatcher as well.  
 *     This is used to implement protocol layering.
 *     Before dispatching, the flow handle will add its own header length
 *     to the dispatch offset.
 * <p>
 * <li>Packet identification.  A flow handle may be used to determine
 *     whether a certain packet belongs to a certain flow.  This is
 *     done by analyzing the packet data.
 * <p>
 * </ul>
 * <p>
 * Uses Linkable for convenience. (XXX)
 *
 * @author	Godmar Back
 * @see FlowDispatcher
 */
public abstract class FlowHandle extends Linkable implements FlowDispatcher
{
    /** synchronizing against that String means excluding irqs */
    static final String	irqlock = "Interrupt Lock";

    private FlowDispatcher	flowDispatcher;
    private LinkedList  	freeBuffers;
    private LinkedList  	usedBuffers;

    /**
     * Checks data in buffer.
     * <p>
     * Checks whether the data in that buffer, starting at offset, may 
     * be identified as belonging to that flow.
     * <p>
     * Sets lastHeaderLength to the number of bytes in the header needed
     * for identification.
     */
    public abstract boolean checkBuffer(Buffer b, int offset);

    /**
     * Optimization:
     * flow handles may set this flag to indicate that checkBuffer will
     * always return true, and thus doesn't have to be called.
     */
    boolean	checkAlwaysTrue = false;

    /** 
     * will contain the number of bytes to skip to get to the payload. <br>
     * This must be set by checkBuffer.
     * This is a hack to avoid having checkBuffer return a pair (boolean, int).
     */
    public int lastHeaderLength;

    /**
     * Class constructor
     */
    /* package */ FlowHandle() 
    {
	freeBuffers = new LinkedList();
	usedBuffers = new LinkedList();
    }

    private String name;  
 
    /**
     * Class constructor
     */
    FlowHandle(String name)
    {
        this();
        this.name = name; 
    }  
 
    public String getName()
    { 
        return name;
    } 
 
    public void setName(String name)
    {
        this.name = name;
    }

    /**
     * Return number of packets in receive queue.
     */
    public int getQueueLength()
    {
	return usedBuffers.count();
    }

    /**
     * Sets a dispatcher for that flow.
     * <p>
     * This has the effect that all packets presented to that flow handle
     * will be presented to the payload dispatcher for dispatch.
     *
     * @param flowDispatcher new payload dispatcher for that flow.
     */
    public void setPayloadDispatcher(FlowDispatcher flowDispatcher)
    {
	this.flowDispatcher = flowDispatcher;
    }

    /**
     * Dispatch a given buffer.
     * <p>
     * If free buffers are available, a free buffer is taken from the
     * free queue.  If buf.exchanged is false, this buffer is returned and
     * the buffer that is passed in is enqueued in the receive queue.
     * <p>
     * Otherwise, the free buffer is made a dependent of the buffer passed
     * in, and it is enqueued in the receive queue.
     * <p>
     * If a payload dispatcher is given, the packet is dispatched to it,
     * adding lastHeaderLength to the offset.
     * <p>
     * See FlowDispatch for a description of the parameters.
     *
     * @see FlowDispatcher#dispatchBuffer
     */
    public Buffer dispatchBuffer(Buffer buf, int offset)
    {
	Buffer exchangebuf = null, ourbuf = null;

	if (!freeBuffers.isEmpty()) {
	    ourbuf = (Buffer)freeBuffers.first;
	    freeBuffers.remove(ourbuf);

	    /* this buffer is already enqueued in somebody else's 
	     * received queue 
	     */
	    if (buf.exchanged) {
		/* make another one */
		ourbuf.makeDependentOf(buf);
	    } else {
		/* otherwise, it's the master, offer our buffer 
		   in exchange */
		exchangebuf = ourbuf;
		ourbuf = buf;
		buf.exchanged = true;
	    }

	    ourbuf.advanceStartBy(offset);

	    /* notify waiters */
	    synchronized (usedBuffers) {
		usedBuffers.append(ourbuf);
		usedBuffers.notify();
	    }
	} else {
	    System.out.println("dropping packet: " + this);
	}

	if (flowDispatcher != null) {
	    Buffer exb = flowDispatcher.dispatchBuffer(buf, 
				offset + lastHeaderLength);
	    if (exb != null)
		exchangebuf = exb;
	}

	return exchangebuf;
    }

    /**
     * Adds a buffer to the handle's receive queue.
     */
    public void addBufferToReceiveQueue(Buffer p)
    {
	synchronized (irqlock) {
	    synchronized (usedBuffers) {
		usedBuffers.append(p);
		usedBuffers.notify();
	    }
	}
    }

    /**
     * Adds a buffer to the handle's free list.
     */
    public void addBuffer(Buffer p)
    {
	synchronized (irqlock) {
	    freeBuffers.append(p.reset());
	}
    }

    /**
     * Adds an array of buffers to the handle's free list.
     */
    public void addBuffers(Buffer p[])
    {
	synchronized (irqlock) {
	    for (int i = 0; i < p.length; i++)
		freeBuffers.append(p[i].reset());
	}
    }

    /**
     * Allocates buffers and adds them to the free list.
     * 
     * @param packets 	number of packets to be allocated
     * @param size 	size of each packet in bytes
     */
    public void addBuffers(int packets, int size)
    {
	synchronized (irqlock) {
	    for (int i = 0; i < packets; i++)
		freeBuffers.append(new Buffer(new byte[size]));
	}
    }

    /**
     * Reclaims a buffer from the free queue.
     *
     * This removes a buffer from the free queue of the handle, effectively
     * reducing the flow handle's capacity.
     *
     * @return free buffer
     */
    public Buffer reclaimBuffer()
    {
	Buffer p = null;
	synchronized (irqlock) {
	    if (!freeBuffers.isEmpty()) {
		freeBuffers.remove(p = (Buffer)freeBuffers.first);
		p.reset();
	    }
	}
	return p;
    }

    /**
     * Receives a packet from that flow.
     * <p>
     * The buffer in which the packet is contained is made independent.
     * (XXX) let the application do that, unless there's security concerns.
     */
    public Buffer receive()
    {
	Buffer p;
	synchronized(usedBuffers) {
	    while (usedBuffers.isEmpty())
		try {
		    usedBuffers.wait();
		} catch (InterruptedException e) {}

	    p = (Buffer)usedBuffers.first;
	    usedBuffers.remove(p);

	    /* may cause a copy */
	    if (!p.isIndependent())
		p.makeIndependent();
	}
	return p;
    }

    /**
     * Merge two flow handles into one
     *
     * This function appends the other flow handle's receive queue to
     * our receive queue.
     */
    public void merge(FlowHandle other)
    {
	synchronized(usedBuffers) {
	    synchronized(other.usedBuffers) {
		while (!other.usedBuffers.isEmpty()) {
		    Buffer p = (Buffer)other.usedBuffers.first;
		    other.usedBuffers.remove(p);
		    usedBuffers.append(p);
		}
	    }
	}
    }

    /**
     * Convert to a String.
     *
     * @return statistics about that flow handle
     */
    public String toString()
    {
	int avail = freeBuffers.count();
	int used = usedBuffers.count();

	return "[free=" + avail + "/used=" + used + "]";
    }
}
