/* 
 *  $Id: KeyArgs.Java,v 1.1.1.1 1998/02/24 04:07:37 gback Exp $
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */


package utils;

class KeyArg {
  String key;
  String arg;

  KeyArg(String k, String a) { key = k; arg = a; }
}


/**
 * Keyword argument manipulation utilities
 *
 * @author David Wetherall
 */


public class KeyArgs {
  private KeyArg[] args;
  private static KeyArg empty = new KeyArg("", null);

  public int length() { return args.length; } 
  public String key(int i) { return args[i].key; }
  public String arg(int i) { return args[i].arg; }

  public String get(String key) {
    for (int i = 0; i < args.length; i++)
      if (args[i].key.equals(key))
	return args[i].arg;
    return null;
  }

  public void set(String key, String arg) {
    for (int i = 0; i < args.length; i++)
      if (args[i].key.equals(key)) {
	args[i].arg = arg;
	return;
      }
    KeyArg[] ret = new KeyArg[args.length + 1];  
    for (int j = 0; j < args.length; j++)
      ret[j] = args[j];

    ret[args.length] = new KeyArg(key, arg);
    args = ret;
  }
  
  public void strike(int i) { args[i] = empty; }

  public void compact() {
    int len = 0;
    for (int i = 0; i < args.length; i++)
      if (args[i] != empty) len++;

    if (len == args.length) return;

    KeyArg[] ret = new KeyArg[len];
    int idx = 0;
    for (int i = 0; i < args.length; i++)
      if (args[i] != empty)
	ret[idx++] = args[i];

    args = ret;
  }
  
  public void merge(String[] ka) 
    throws Exception
  {
    merge(new KeyArgs(ka));
  }

  public void merge(String[] ka, int skip) 
    throws Exception
  {
    merge(new KeyArgs(ka, skip));
  }

  public void merge(KeyArgs k) {
    compact();
    k.compact();
	
    int len = args.length;
    for (int i = 0; i < k.args.length; i++) {
      boolean match = false;
      for (int j = 0; j < args.length; j++)
	if (args[j].key.equals(k.args[i].key)) {
	  match = true;
	  break;
	}
      if (!match) len++;
    }
    
    KeyArg[] ret = new KeyArg[len];  
    for (int j = 0; j < args.length; j++)
      ret[j] = args[j];
    
    int idx = args.length;
    for (int i = 0; i < k.args.length; i++) {
      boolean match = false;
      for (int j = 0; j < args.length; j++)
	if (args[j].key.equals(k.args[i].key)) {
	  match = true;
	  break;
	}
      if (!match) ret[idx++] = k.args[i];
    }

    args = ret;
  }

  public String toString() {
    compact();
    String ka = "";
    for (int i = 0; i < args.length; i++) {
      ka += "\"" + args[i].key + "\" ";
      if (args[i].arg != null) ka += "\"" + args[i].arg + "\" ";
    }
    return ka;
  }
  
  public String[] toStrings() {
    compact();
    int len = args.length;
    for (int i = 0; i < args.length; i++)
      if (args[i].arg != null) len++;

    String[] ret = new String[len];

    int idx = 0;
    for (int i = 0; i < args.length; i++) {
      ret[idx++] = args[i].key;
      if (args[i].arg != null)
	ret[idx++] = args[i].arg;
    }

    return ret;
  }

  private KeyArgs(KeyArg[] ka) {
    args = ka;
  }
  
  public KeyArgs(String[] ka)
    throws Exception
  {
    this(ka, 0);
  }

  public KeyArgs() 
    throws Exception
  {
    this(null, 0);
  }

  public KeyArgs(String[] ka, int skip)
    throws Exception
  {
    // handle the trivial case first
    if (ka == null || ka.length == 0 || ka.length <= skip) {
      args = new KeyArg[0];
      return;
    }
    
    // walk the list once, counting and checking syntax
    int l = 0;
    boolean ok = false;
    for (int i = skip; i < ka.length; i++) {
      if (ka[i] != null && ka[i].length() > 0 && ka[i].charAt(0) == '-') {
	l++;
	ok = true;
	continue;
      }
      if (!ok) {
	String ks = "";
	for (int j = 0; j < ka.length; j++)
	  ks += ka[i] + " ";
	throw new Exception("invalid keyags sequence: " + ks);
      }
      ok = false;
    }
    
    // now fill it in for real 
    args = new KeyArg[l];
    int idx = 0;
    for (int i = skip; i < ka.length; i++) {
      if (ka[i] != null && ka[i].length() > 0 &&  ka[i].charAt(0) == '-')
	args[idx++] = new KeyArg(ka[i], null);
      else
	args[idx-1].arg = ka[i];
    }
  }


  /** Test the class. */
  public static void main(String[] args) 
    throws Exception
  {
    String[] defaults = { "-one", "one" };
    String[] ka = { "-one", "1", "-two", "-three", "3" };
    String[] kb = { "-four" };
    String[] kc = { "-one", "1", "bogus" }; 

    KeyArgs k = new KeyArgs(ka);
    if (k.length() != 3) throw new Exception("wrong initial length");

    k.merge(defaults);
    if (k.length() != 3) throw new Exception("wrong unchanged merge length");

    String[] ks = k.toStrings();
    if (ka == null || ks == null || ka.length != ks.length)
      throw new Exception("representations really differ");
    for (int i = 0; i < ka.length; i++)
      if (!ka[i].equals(ks[i])) 
	throw new Exception("representations differ");
    
    k.merge(kb);
    if (k.length() != 4) throw new Exception("wrong expanded merge length");

    if (!k.get("-one").equals("1")) throw new Exception("wrong get value");

    k.set("-five", null);
    if (k.length() != 5) throw new Exception("wrong expanded set length");

    k.set("-five", "5");
    if (!k.get("-five").equals("5")) throw new Exception("wrong set value");

    k.strike(0);
    if (k.length() != 5) throw new Exception("wrong strike length");

    k.compact();
    if (k.length() != 4) throw new Exception("wrong compact length");
    
    k.strike(0);
    int ml = k.length() - 1;
    k.merge(k);
    if (k.length() != ml) throw new Exception("wrong compacted merge length");

    try {
      KeyArgs x = new KeyArgs(kc);
      throw new Exception("invalid");
    } catch (Exception e) {
      if (e.getMessage().equals("invalid"))
	throw new Exception("invalid list accepted");  
    }
  }
}
