/* 
 *  $Id: LineListTokenizer.Java,v 1.4 1998/06/01 04:50:28 gback Exp $
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */


package utils;

import java.io.*;


/**
 * File reading utility -- converts each non-empty line to a word list
 *
 * @author David Wetherall
 */


public class LineListTokenizer
  extends StreamTokenizer 
{

  // sorry, no more than 256 elements per line
  final static int MAXLIST = 256;
  String[] list = new String[MAXLIST];
  boolean done;
  int length;


  /**
   * Parse the next non-empty line into an array of strings.
   * Result is null when EOF, and otherwise has at least one element.
   */

  public String[] list() 
    throws Exception 
  {
    length = 0;
    
    while (!done && length == 0) {
      do {
	int t = nextToken();
	
	if (t == TT_EOF)
	  done = true;
	else if (t == TT_EOL)
	  break;
	// numbers not expected but handle them anyhow
	else if (t == TT_NUMBER)
	  list[length++] = Double.toString(nval);
	else
	  list[length++] = sval;
      } while (!done && length < MAXLIST);
    }
    
    if (length == MAXLIST) throw new Exception("Line too long");
    if (done || length == 0) return null;

    String[] result = new String[length];
    System.arraycopy(list, 0, result, 0, length);
    return result;
  }


  public LineListTokenizer(String file)
    throws FileNotFoundException
  {
    this(new InputStreamReader(new FileInputStream(file)));
  }

  public LineListTokenizer(Reader input)
  {
    super(input);

    resetSyntax();
    whitespaceChars(0, ' ');
    wordChars('0', '9');
    wordChars('-', '-');
    wordChars('.', '.');
    wordChars(':', ':');
    wordChars('=', '=');
    wordChars('/', '/');
    wordChars('A', 'Z');
    wordChars('a', 'z');
    commentChar('#');
    quoteChar('\'');
    quoteChar('\"');    
    eolIsSignificant(true);
  }


  /** Test the class. */
  public static void main(String[] args)
    throws Exception
  {
    String file = (args.length > 0) ? args[0] : "tmp";

    PrintWriter p = new PrintWriter(new FileOutputStream(file));
    p.println("#");
    p.println("one");
    p.println(" two two ");
    p.println("1 2  3");
    p.println("# 1 2  3");
    p.println("four four 4 4");
    p.println("");
    p.println("5 5 5 5 5");
    p.println(" ");
    p.close();

    LineListTokenizer lt = new LineListTokenizer(file);
    String[] line;
    
    // expected output is 1 -> 5
    while ((line = lt.list()) != null) {
      System.out.println(line.length);
    }

    Runtime.getRuntime().exec("/bin/rm " + file);
  }
}
